@@ -59,6 +59,7 @@ int openssl_in_fips_mode(void)
 #define HMAC_SUFFIX ".hmac"
 #define READ_BUFFER_LENGTH 16384
 
+#ifndef ICA_INTERNAL_TEST
 /*
  * The hard-coded HMAC key to be optionally provided for the library
  * integrity test. The recommended key size for HMAC-SHA256 is 64 bytes.
@@ -69,6 +70,8 @@ static const char hmackey[] =
 	"0000000000000000000000000000000000000000000000000000000000000000"
 	"0000000000000000000000000000000000000000000000000000000000000000";
 
+#endif /* ICA_INTERNAL_TEST */
+
 int fips;
 
 #define LIBICA_FIPS_CONFIG		LIBICA_CONFDIR "/libica/openssl3-fips.cnf"
@@ -253,10 +256,23 @@ fips_init(void)
 		}
 
 		fips |= ICA_FIPS_MODE;
+#endif
+	} else {
+		/* kernel fips flag == 0, load default provider in case we are
+		 * running with openssl 3.0 */
+#if OPENSSL_VERSION_PREREQ(3, 0)
+		openssl_provider = OSSL_PROVIDER_load(openssl_libctx, "default");
+		if (openssl_provider == NULL) {
+			syslog(LOG_ERR, "Libica: failed to load default provider\n");
+			fips |= ICA_FIPS_INTEGRITY;
+			return;
+		}
 #endif
 	}
 }
 
+#ifndef ICA_INTERNAL_TEST
+
 static char *make_hmac_path(const char *origpath)
 {
 	char *path;
@@ -317,6 +333,7 @@ static int compute_file_hmac(const char *path, void **buf, size_t *hmaclen)
 		}
 	}
 
+	hlen = sizeof(rbuf);
 	if (EVP_DigestSignFinal(mdctx, rbuf, &hlen) <= 0)
 		goto end;
 
@@ -343,6 +360,37 @@ end:
 	return rc;
 }
 
+static int load_known_hmac(const char *path, unsigned char **hmac, long *hmaclen)
+{
+	int rc = -1;
+	FILE *fp;
+	char *known_hmac_str = NULL;
+	char *hmacpath, *p;
+	size_t n;
+
+	hmacpath = make_hmac_path(path);
+	if (hmacpath == NULL)
+		return rc;
+
+	if ((fp = fopen(hmacpath, "r")) == NULL)
+		return rc;
+
+	if (getline(&known_hmac_str, &n, fp) <= 0)
+		goto end;
+
+	if ((p = strchr(known_hmac_str, '\n')) != NULL)
+		*p = '\0';
+
+	*hmac = OPENSSL_hexstr2buf(known_hmac_str, hmaclen);
+	rc = 0;
+end:
+	fclose(fp);
+	free(known_hmac_str);
+	free(hmacpath);
+
+	return rc;
+}
+
 /**
  * Performs the FIPS check.
  *
@@ -352,50 +400,28 @@ end:
 static int FIPSCHECK_verify(const char *path)
 {
 	int rc = 0;
-	FILE *fp;
 	unsigned char *known_hmac = NULL;
-	long hmaclen;
-	char *hmacpath, *p;
-	char *known_hmac_str = NULL;
-	size_t n, computed_hmac_len;
+	long known_hmac_len;
 	void *computed_hmac = NULL;
+	size_t computed_hmac_len;
 
-	hmacpath = make_hmac_path(path);
-	if (hmacpath == NULL)
-		return 0;
-
-	fp = fopen(hmacpath, "r");
-	if (fp == NULL) {
+	if (load_known_hmac(path, &known_hmac,  &known_hmac_len) != 0)
 		goto end;
-	}
-
-	if (getline(&known_hmac_str, &n, fp) <= 0)
-		goto end;
-
-	if ((p = strchr(known_hmac_str, '\n')) != NULL)
-		*p = '\0';
-
-	known_hmac = OPENSSL_hexstr2buf(known_hmac_str, &hmaclen);
 
 	if (compute_file_hmac(path, &computed_hmac, &computed_hmac_len) != 0)
 		goto end;
 
+	if ((size_t)known_hmac_len != computed_hmac_len)
+		goto end;
+
 	if (memcmp(computed_hmac, known_hmac, computed_hmac_len) != 0)
 		goto end;
 
 	rc = 1;
-
 end:
-
 	free(computed_hmac);
-	free(known_hmac_str);
-	free(hmacpath);
-
 	OPENSSL_free(known_hmac);
 
-	if (fp)
-		fclose(fp);
-
 	return rc;
 }
 
@@ -469,6 +495,7 @@ static void fips_lib_integrity_check(void)
 
 	syslog(LOG_INFO, msg3);
 }
+#endif /* ICA_INTERNAL_TEST */
 
 void
 fips_powerup_tests(void)
@@ -1058,7 +1085,7 @@ _err_:
 }
 
 static int
-des3_cbc_cs_kat(void){
+des3_cbc_cs_kat(void) {
 	const struct des3_cbc_cs_tv *tv;
 	size_t i;
 	unsigned char iv[DES3_BLKSIZE], *out;
@@ -1287,11 +1314,34 @@ rsa_kat(void)
 	ica_adapter_handle_t ah;
 	const struct rsa_tv *tv;
 	size_t i, keylen, crtparamlen;
-	unsigned char *out;
+	unsigned char *out = NULL;
+	libica_func_list_element* libica_func_list = NULL;
+	unsigned int count;
 
 	if (ica_open_adapter(&ah))
 		return 1;
 
+	if (ica_get_functionlist(NULL, &count) != 0)
+		goto _err_;
+
+	libica_func_list = malloc(sizeof(libica_func_list_element) * count);
+	if (!libica_func_list)
+		goto _err_;
+
+	if (ica_get_functionlist(libica_func_list, &count) != 0)
+		goto _err_;
+
+	for (i = 0; i < count; i++) {
+		if (libica_func_list[i].mech_mode_id == RSA_CRT &&
+			libica_func_list[i].flags == 0) {
+			/* RSA_CRT, and probably also RSA_ME, not available, skip test.
+			 * Looks like we don't have cards nor sw fallbacks. */
+			free(libica_func_list);
+			ica_close_adapter(ah);
+			return 0;
+		}
+	}
+
 	for (i = 0; i < RSA_TV_LEN; i++) {
 		tv = &RSA_TV[i];
 
@@ -1340,10 +1390,12 @@ rsa_kat(void)
 		free(privkey.dq);
 		free(privkey.qInverse);
 	}
+	free(libica_func_list);
 	ica_close_adapter(ah);
 	return 0;
 
 _err_:
+	free(libica_func_list);
 	ica_close_adapter(ah);
 	free(out);
 	free(pubkey.exponent);
